#ifndef CALENDAR_HPP
#define CALENDAR_HPP


enum DayOfWeek
{
	SUNDAY,
	MONDAY,
	TUESDAY,
	WEDNESDAY,
	THURSDAY,
	FRIDAY,
	SATURDAY
};

enum Calendar
{
	GREGORIAN,
	JULIAN,
	REFORMED_JULIAN
};

inline long int Quotient(long int a, long int b, long int &remainder)
{
	long int q, r;
	
	if (b==0)
		throw("Dividing by zero");
	
	q = a/b;
	r = a - q*b;
	
	if (r<0)
	{
		if (b>0)
		{
			r += b;
			q--;
		}
		else
		{
			r -= b;
			q++;
		}
	}
	
	remainder = r;
	
	return q;
}

inline long int Quotient(long int a, long int b)
{
	long int dummy;
	
	return Quotient(a, b, dummy);
}

inline long int Remainder(long int a, long int b)
{
	long int remainder;
	
	Quotient(a, b, remainder);
	
	return remainder;
}

inline bool IsLeapYear(long int year, Calendar calendar = GREGORIAN)
{
	if (Remainder(year, 4) != 0)
		return false;
	
	if (calendar == GREGORIAN)
		return Remainder(year, 100)!=0 || Remainder(year, 400)==0;
	
	if (calendar == JULIAN)
		return true;
	
	if (calendar == REFORMED_JULIAN)
		return Remainder(year, 100)!=0 || Remainder(year, 900)%400==200;
	
	throw("Unrecognized calendar type");
}

inline void OrdinaryToModified(long int &year, int &month)
{
	if (month>=3)
		month -= 3;
	else
	{
		month += 9;
		year--;
	}
}

inline void ModifiedToOrdinary(long int &year, int &month)
{
	if (month<=9)
		month += 3;
	else
	{
		month -= 9;
		year++;
	}
}

inline int DaysInPreviousMonths(int modifiedMonth)
{
	return (153*modifiedMonth + 2)/5;
}

inline bool IsValidDate(long int year, int month, int day, Calendar calendar = GREGORIAN)
{
	if (calendar!=GREGORIAN && calendar!=JULIAN && calendar!=REFORMED_JULIAN)
		throw("Unrecognized calendar type");
	
	if (month<1 || month>12)
		return false;
	
	if (month == 2)
		if (day<1 || day>29)
			return false;
		else
			if (day != 29)
				return true;
			else
				return IsLeapYear(year, calendar);
	
	if (day<1)
		return false;
	
	OrdinaryToModified(year, month);
	
	if (day>DaysInPreviousMonths(month+1)-DaysInPreviousMonths(month))
		return false;
	
	return true;
}

inline void CheckValidity(long int year, int month, int day, Calendar calendar = GREGORIAN)
{
	if (!IsValidDate(year, month, day, calendar))
		throw("Invalid date");
}

inline void DateFromYearAndDayNumberInYear(int dayNumberInYear, long int &year, int &month, int &day)
{
	if (dayNumberInYear<1 || dayNumberInYear>366)
		throw("Day number in year must be in range 1-366");
	
	month = (5*dayNumberInYear - 3)/153;
	day = dayNumberInYear - DaysInPreviousMonths(month);
	ModifiedToOrdinary(year, month);
}

inline long int JulianDateToNumber(long int year, int month, int day)
{
	CheckValidity(year, month, day, JULIAN);
	OrdinaryToModified(year, month);
		
	return day + DaysInPreviousMonths(month) + 365*year + Quotient(year, 4);
}

inline long int GregorianDateToNumber(long int year, int month, int day)
{
	CheckValidity(year, month, day, GREGORIAN);
	OrdinaryToModified(year, month);
	
	return day + DaysInPreviousMonths(month) + 365*year + Quotient(year, 4) - Quotient(year, 100) + Quotient(year, 400) + 2;	
}

inline long int ReformedJulianDateToNumber(long int year, int month, int day)
{
	CheckValidity(year, month, day, REFORMED_JULIAN);
	OrdinaryToModified(year, month);
	
	return day + DaysInPreviousMonths(month) + 365*year + Quotient(year, 4) - Quotient(year, 100) + Quotient(Remainder(year, 900) + 200, 400) + 6;
}

inline void JulianDateFromNumber(long int number, long int &year, int &month, int &day)
{
	year = 4*Quotient(number, 365*4+1, number);
	year += Quotient(number, 365, number);
	DateFromYearAndDayNumberInYear(number, year, month, day);
}

inline void GregorianDateFromNumber(long int number, long int &year, int &month, int &day)
{
	number -= 2;
	year = 400*Quotient(number, 365*400 + 400/4 - 3, number);
	year += 100*Quotient(number, 365*100 + 100/4 - 1, number);
	year += 4*Quotient(number, 365*4+1, number);
	year += Quotient(number, 365, number);
	DateFromYearAndDayNumberInYear(number, year, month, day);
}

inline void ReformedJulianDateFromNumber(long int number, long int &year, int &month, int &day)
{
	number -= 2;
	year = 900*Quotient(number, 365*900 + 900/4 - 7, number);
	year += 400*Quotient(number + 365*200 + 48, 365*400 + 400/4 - 3, number);
	number -= 365*200 + 48;
	year += 100*Quotient(number, 365*100 + 24, number);
	year += 4*Quotient(number, 365*4+1, number);
	year += Quotient(number, 365, number);
	DateFromYearAndDayNumberInYear(number, year, month, day);
}

inline long int DateToNumber(long int year, int month, int day, Calendar calendar = GREGORIAN)
{
	CheckValidity(year, month, day, calendar);
	
	if (calendar == GREGORIAN)
		return GregorianDateToNumber(year, month, day);
	else if (calendar == JULIAN)
		return JulianDateToNumber(year, month, day);	
	else if (calendar == REFORMED_JULIAN)
		return ReformedJulianDateToNumber(year, month, day);
	
	throw("Unrecognizing calendar type is unexpected. This is a bug!");
}

inline void DateFromNumber(long int number, long int &year, int &month, int &day, Calendar calendar = GREGORIAN)
{
	if (calendar == GREGORIAN)
		GregorianDateFromNumber(number, year, month, day);
	else if (calendar == JULIAN)
		JulianDateFromNumber(number, year, month, day);	
	else if (calendar == REFORMED_JULIAN)
		ReformedJulianDateFromNumber(number, year, month, day);
	else
		throw("Unknown calendar type");
}

inline DayOfWeek GetDayOfWeek(long int dayNumber)
{
	return (DayOfWeek) ((dayNumber)%7);
}

inline DayOfWeek GetDayOfWeek(long int year, int month, int day, Calendar calendar = GREGORIAN)
{
	return GetDayOfWeek(DateToNumber(year, month, day, calendar));
}

inline long int EasterDayNumberFromJulianYear(long int year)
{
	long int day, month, remainder;
	
	Quotient(year, 19, day);
	Quotient(19*day + 15, 30, day);
	Quotient(day + year + Quotient(year, 4), 7, remainder);
	day += 28 - remainder;

	if (day>31)
	{
		day -= 31;
		month = 4;
	}
	else
		month = 3;
	
	return DateToNumber(year, month, day, JULIAN);
}

inline long int EasterDayNumber(long int year, Calendar calendar = GREGORIAN)
{
	long int begin, end, dayNumber;
	int day, month;
	
	begin = DateToNumber(year, 1, 1, calendar);
	end = DateToNumber(year + 1, 1, 1, calendar);
	DateFromNumber(begin, year, month, day);
	
	while (true)
	{
		dayNumber = EasterDayNumberFromJulianYear(year);
		
		if (dayNumber<begin)
			break;
		
		year--;
	}
	
	while (true)
	{
		year++;
		dayNumber = EasterDayNumberFromJulianYear(year);
		
		if (dayNumber>=begin)
			break;
	}
	
	if (dayNumber>=end)
		throw("No ortodox Easter in this year");
	
	return dayNumber;
}

inline void Easter(long int &year, int &month, int &day, Calendar calendar = GREGORIAN)
{
	long int dayNumber;
	
	dayNumber = EasterDayNumber(year, calendar);
	DateFromNumber(dayNumber, year, month, day);
}

inline long int GetDayNumberOfDayOfWeekFrom(long int number, DayOfWeek dayOfWeek)
{
	return number + Remainder(dayOfWeek - number, 7);
}

inline long int GetDayNumberOfDayOfWeekTo(long int number, DayOfWeek dayOfWeek)
{
	return number - Remainder(number - dayOfWeek, 7);
}

inline long int GetDayNumberOfNextDayOfWeek(long int number, DayOfWeek dayOfWeek)
{
	return number - Remainder(number - dayOfWeek, 7) + 7;
}

inline long int GetDayNumberOfPreviousDayOfWeek(long int number, DayOfWeek dayOfWeek)
{
	return number + Remainder(dayOfWeek - number, 7) - 7;
}

inline void DayOfWeekFrom(DayOfWeek dayOfWeek, long int &year, int &month, int &day, Calendar calendar = GREGORIAN)
{
	CheckValidity(year, month, day, calendar);
	DateFromNumber(GetDayNumberOfDayOfWeekFrom(DateToNumber(year, month, day, calendar), dayOfWeek), year, month, day, calendar);
}

inline void DayOfWeekTo(DayOfWeek dayOfWeek, long int &year, int &month, int &day, Calendar calendar = GREGORIAN)
{
	CheckValidity(year, month, day, calendar);
	DateFromNumber(GetDayNumberOfDayOfWeekTo(DateToNumber(year, month, day, calendar), dayOfWeek), year, month, day, calendar);
}

inline void NextDayOfWeek(DayOfWeek dayOfWeek, long int &year, int &month, int &day, Calendar calendar = GREGORIAN)
{
	CheckValidity(year, month, day, calendar);
	DateFromNumber(GetDayNumberOfNextDayOfWeek(DateToNumber(year, month, day, calendar), dayOfWeek), year, month, day, calendar);
}

inline void PreviousDayOfWeek(DayOfWeek dayOfWeek, long int &year, int &month, int &day, Calendar calendar = GREGORIAN)
{
	CheckValidity(year, month, day, calendar);
	DateFromNumber(GetDayNumberOfPreviousDayOfWeek(DateToNumber(year, month, day, calendar), dayOfWeek), year, month, day, calendar);
}


#endif
